#define _USE_MATH_DEFINES
#include <iostream>
#include <complex>
#include <vector>
#include <math.h>
#include <stdlib.h>
#include "chart.h"

using namespace std;
typedef complex<double> cplx_num;   // Eine einzelne komplexe Zahl
typedef vector<cplx_num> cplx_vec;  // Vektor mit komplexen Zahlen

// ---------------------------------------------------------------------------
// FFT für Vektor a, gefüllt mit komplexen Werten, berechnen.
// Achtung: Die Anzahl der Elemente im Vektor a muss eine Zweierpotenz sein!
// ---------------------------------------------------------------------------
cplx_vec fft(cplx_vec a)
{
    // Abbruch der Rekursion?
    int n = a.size();
    if(n < 2) return a;

    cplx_num I(0, 1);  // Imaginäre Einheit
    cplx_num omega = 1;
    cplx_num omega_n = exp(-2 * M_PI / n * I);

    // Teilvektoren a_i und a_ii füllen
    cplx_vec a_i, a_ii;
    for(auto x : a)
    {
        if(a_i.size() <= a_ii.size())
            a_i.push_back(x);
        else
            a_ii.push_back(x);
    }

    // Rekursiver Selbstaufruf
    auto y_i  = fft(a_i);
    auto y_ii = fft(a_ii);

    // Neuen Vektor y mit n Elementen anlegen und füllen
    cplx_vec y(n);
    for(int k = 0; k < n / 2; k++)
    {
        y[k] = y_i[k] + omega * y_ii[k];
        y[k + n / 2] = y_i[k] - omega * y_ii[k];
        omega = omega * omega_n;
    }

    return y;
}

// ---------------------------------------------------------------------------
// Folgendes "Hauptprogramm" führt einen ersten Test der FFT-Funktion durch.
// ---------------------------------------------------------------------------
int main(void)
{
    // Vektor mit 8 Elementen, die alle den Wert 1 haben
    cplx_vec y(8, 1);
    auto Y = fft(y);

    // Das erste Element ist 8, die anderen Elemente sind 0
    for(auto x : Y) cout << x << endl;

    /*
        % Die entsprechenden MATLAB-Befehle zum Vergleich:
        y = ones(1, 8);
        Y = fft(y);
        disp(Y);
    */

    return 0;
}

// ---------------------------------------------------------------------------
// Dieses Hauptprogramm wurde aus der MATLAB-Dokumentation ("doc fft")
// zur Funktion fft übernommen. Die entsprechenden MATLAB-Befehle sind
// als Kommentare im Quelltext angegeben (s. u.)...
// ---------------------------------------------------------------------------
int main1(void)
{
    /*
        Fs = 1000;                    % Sampling frequency
        T = 1/Fs;                     % Sample time
        L = 1000;                     % Length of signal
        t = (0:L-1)*T;                % Time vector

        % Sum of a 50 Hz sinusoid and a 120 Hz sinusoid
        x = 0.7*sin(2*pi*50*t) + sin(2*pi*120*t);
        y = x + 2*randn(size(t));     % Sinusoids plus noise
    */

    cplx_vec y;
    double Fs = 1000, T = 1.0 / Fs, L = 1024, t = 0;
    while(y.size() < L)
    {
        double x = 0.7 * sin(2 * M_PI * 50 * t) + sin(2 * M_PI * 120 * t);
        double y_ = x + 2.0 * rand() / RAND_MAX - 1.0;
        chart_lineto(t, y_, CHART_RED);
        y.push_back(y_); t = t + T;
    }

    /*
        plot(Fs*t(1:50),y(1:50))
        title('Signal Corrupted with Zero-Mean Random Noise')
        xlabel('time (milliseconds)')
    */

    chart_axis_fmt("%g ms", "%g");
    chart_show(CHART_YZOOM_OFF, "Signal im Zeitbereich");

    /*
        NFFT = 2^nextpow2(L); % Next power of 2 from length of y
        Y = fft(y,NFFT)/L;
        f = Fs/2*linspace(0,1,NFFT/2+1);
    */

    cplx_vec Y = fft(y);

    /*
        % Plot single-sided amplitude spectrum.
        plot(f,2*abs(Y(1:NFFT/2+1)))
        title('Single-Sided Amplitude Spectrum of y(t)')
        xlabel('Frequency (Hz)')
        ylabel('|Y(f)|')
    */

    chart_clear();
    for(int idx = 0; idx <= L / 2; idx++)
        chart_lineto(Fs * idx / L, 2.0 * abs(Y[idx]) / L, CHART_BLUE);
    chart_axis_fmt("%g Hz", "%g");
    chart_show(CHART_YZOOM_OFF, "Amplitudenspektrum");
    return 0;
}
